/* EDIT - routines to implement single line entry field and multi-line
	entry field. */


/* ------------------------ */
#include "getgui.h"


/* Keys.. */
#define LEFT 551	/* left arrow */
#define UP 552		/* up arrow */
#define RIGHT 553	/* right arrow */
#define DOWN 554	/* down arrow */
#define HOME 550	/* HOME */
#define CTRL_A 1 	/* same as HOME */
#define END  557 	/* END */
#define CTRL_E 5	/* same as END */
#define CTRL_K 11	/* delete to end of line */
#define CTRL_U 21	/* delete entire line */
#define DEL '\177'	/* DEL */
#define CTRL_D	4	/* ^D */
#define BACKSPACE 8	/* ^H */


/* the following may be set to 1 to make DEL=BACKSPACE, 0 if not */
#define DEL_EQUIV_BACKSPACE 0

/* appearance */
#define MARG 0.08

/* # lines allowed in a multi-line editor */
#define MAXLINES 40
 


static int addbuf(), delbuf(), delword(), stp(), cursor(), rubout();
static int stpp(), refresh();

/* even though this editln() is potentially recursive (via expose events)
   the following are safe because they are set upon each entry */
static double x, y, height;
static char bkcolor[20];

 
static int Break[MAXLINES];
static int Nlines;
static int Alignerrflag;



/* ------------------- */
/* EDITLN - a single field for entering/editing text. 
   Returns: the terminating key, or 0 when in Display mode.
   Notes: desired text size must be set before call.
*/
getgui_editln( mode, buf, x1, y1, len, pos, modflag )
int mode; /* either 'N', 'E', 'D', 'R', or 'P', 'X'  
		  (New, Edit, Display, Redraw, Password, Password redraw) */
char *buf;  /* the text */
double x1, y1;	/* (x,y) of lower left corner of entry area */
int len;    /* max length of box in characters */
int *pos;   /* current cursor position (updated here) */
int *modflag; /* 1 if text was modified, 0 if not */
{

int	i, j;
int	r, c;		/* current row and column of cursor */
int	key; 		/* most recent key pressed (extended are in the 500's range) */
double width;
double ex, ey;
int dlindex;
char *str;		/* the stuff being typed */
int textsize;
char s[4];
int permmode;


*modflag = 0;
if( len > 126 ) len = 126;

if( mode == 'N' ) buf[0] = '\0';
if( mode == 'D' ) strcpy( bkcolor, Color3 );
else strcpy( bkcolor, Color4 );

/* draw area */
textsize = Ecurtextsize; /* remember in case it changes underfoot */
x = x1;
y = y1;
width = len * Ecurtextwidth;
height = Ecurtextheight;
Ecblock( x-MARG, y-MARG, x+width+MARG, y+height+MARG, bkcolor, 0 );
Ecblockdress( x-MARG, y-MARG, x+width+MARG, y+height+MARG, 
	0.02, "gray(1.0)", "gray(0.3)", 0.0, "" );


/* display current contents if any.. */
stp( 0 ); 
if( mode != 'P' && mode != 'X' ) Etext( buf );

/* save info in display list (use 'permanent' mode) */
if( mode == 'R' ) permmode = 'E';
else if( mode == 'X' ) permmode = 'P';
else permmode = mode;
if( Keepdl ) dlindex = getgui_dladd( "E", buf, x-MARG, y-MARG, x+width+MARG, y+height+MARG, x, y, 
				permmode, len, *pos, Ecurtextsize, 0 );

/* find this editln's location in the display list */
else dlindex = getgui_dlfind( x, y );


/* initial cursor position.. */
if( *pos >= strlen( buf )) c = strlen( buf );
else c = *pos;
if( mode != 'D' ) cursor( c, 1 );

if( mode == 'D' || mode == 'R' || mode == 'X' ) return( 0 ); /* redraws return here.. */

/* work directly on entry in display list, in case an expose occurs in the 
   midst of what we are doing.. */
str = DLval[ dlindex ];


/* begin getting keystrokes.. */
while( 1 ) {

	stp( c );
	DLi3[ dlindex ] = c;
	Egetkey( &ex, &ey, &key );

	/* make sure textsize hasn't changed underfoot by expose/redraw */
	if( Ecurtextsize != textsize ) Etextsize( textsize );


	if( key == LEFT ) {
		if( c > 0 && len > 1 ) { cursor( c, 0 ); c--; cursor( c, 1 ); }
		}
		
	else if( key == RIGHT ) {
		if(  c < ( strlen( str )) ) { cursor( c, 0), c++; cursor( c, 1 ); }
		}
		
	else if( key == HOME || key == CTRL_A ) { cursor( c, 0 ); c = 0; cursor( c, 1 ); } 

	else if( key == END || key == CTRL_E ) { cursor( c, 0 ); c = strlen( str ); cursor(c,1); }

	else if( key == E_MOUSE_LEFT ) { /* move cursor to desired location */
		if( ey < y-MARG || ey > (y + height + MARG) ) break;
		else if( ex < x || ex > x + width ) break;
		else if( ex > x + (strlen(str) * Ecurtextwidth )) {
			cursor( c, 0 ); 
			c = strlen( str );
			cursor( c, 1 );
			continue;
			}
		/* cursor( c, 0 ); */
		c = (int)(((ex - x) / Ecurtextwidth) + 0.8 );
		/* cursor( c, 1 ); */
		}

	else if( key == E_MOUSE_MIDDLE || key == E_MOUSE_RIGHT ) { /* increment/decrement */
		if( ey < y || ey > (ey + height) ) break;
		else if( ex < x || ex > x + width ) break;
		else if( ex > x + (strlen(str) * Ecurtextwidth )) continue;
		j = (int)(((ex - x) / Ecurtextwidth) );
		if( isdigit( str[j] )) {
			if( key == E_MOUSE_MIDDLE ) str[j]--;
			else if( key == E_MOUSE_RIGHT ) str[j]++;
			if( str[j] > '9' ) str[j] = '0';
			if( str[j] < '0' ) str[j] = '9';
			/* rubout( j, 1 );
			 * stp( j );
			 * sprintf( s, "%c", str[j] );
			 * if( mode != 'P' ) Etext( s );
			 * cursor( c, 1 );
			 */
			}
		}

	else if( key == CTRL_K ) {
		/* cursor( c, 0 ); */
		str[c] = '\0';
		/* c = 0; */
		/* cursor( c, 1 );
		 * rubout( 0, len ); */
		*modflag = 1;
		}

	else if( key == CTRL_U ) {
		str[0] = '\0';
		c = 0;
		*modflag = 1;
		}

	else if( key == CTRL_D || key == DEL || key == BACKSPACE ) {
		*modflag = 1;
		if( key == BACKSPACE || (key == DEL && DEL_EQUIV_BACKSPACE) ) { 
			if( c > 0 ) { 
				/* cursor( c, 0 ); */
				c--; 
				/* cursor( c, 1 ); */
				}
			else if( c == 0 ) continue;
			}
		if( c < strlen( str ) ) {
			delbuf( str, c, 1 );
			/* stp( c );
			 * rubout( c, len-c );
			 * stp( c );
			 * if( mode != 'P' ) Etext( &str[c] ); 
			 * rubout( strlen( str ), 1 );
			 * cursor( c, 1 );
			 */
			}
		}

	/* typed characters.. */
	else if( key >= 32 && key <= 126 ) {
		if( strlen( str ) >= len ) continue;
		*modflag = 1;
		addbuf( str, c, key );
		/* stp( c );
		 * rubout( c, len-c );
		 * stp( c );
		 * if( mode != 'P' )Etext( &str[c] ); 
		 * cursor( c, 0 ); 
		 */
		c++; 
		/* cursor( c, 1 ); */
		}

	else 	{
		break; /* and return the pressed key.. */
		}

	/* redraw */
	Ecblock( x-MARG, y-MARG, x+width+MARG, y+height+MARG, bkcolor, 0 );
	stp( 0 ); 
	if( mode != 'P' ) Etext( str );
	cursor( c, 1 );
	}

*pos = c;

strcpy( buf, str );
return( key );
}

/* ============================= */ 
/* DELBUF - Delete some text from the buffer */
/* txt = text buffer */
/* itxt = current position */
/* dels = number of chars to delete */
static int
delbuf( txt, itxt, dels )
char *txt;
int itxt;
int dels;
{
int i;
if( itxt < 0 || itxt > strlen( txt ) ) fprintf( stderr, "delbuf: itxt (%d) out of range\n", itxt );
for( i = itxt; i <= (int) (strlen( txt )) - (int) dels; i++ ) {
	txt[i] = txt[i+dels];
	}
}

/* ============================= */ 
/* ADDBUF - Add some text to the buffer */
/* txt = text buffer */
/* itxt = current position */
/* newchar = char to be added */

static int addbuf( txt, itxt, newchar )
char *txt;
int itxt;
char newchar;
{
int i;
if( itxt < 0 || itxt > strlen( txt ) ) fprintf( stderr, "addbuf: itxt (%d) out of range\n", itxt );
for( i = strlen( txt ); i >= itxt ; i-- ) {
	txt[i + 1] = txt[i];
	}
txt[itxt] = newchar;
}
/* ============================== */
/* DELWORD - delete word which cursor is within.. */
static int delword( buf, c )
char *buf;
int *c;
{
int i, j;

if( buf[*c] == ' ' ) {
	if( *c > 0 && buf[*c-1] != ' ' ) (*c)--;
	else return( 0 );
	}

/* find beginning of word.. */
for( i = *c; i > 0; i-- ) if( buf[i] == ' ' ) break;

/* find end of word.. */
for( j = *c; j < strlen( buf ); j++ ) if( buf[j] == ' ' ) break;

strcpy( &buf[i], &buf[j] );
*c = i;
return( 0 );
}

/* ---------------- */
/* stp - set current text position. */
static int
stp( col )
int col;
{
Emov( x + (col * Ecurtextwidth), y );
return( 0 );
}

/* ------------------- */
/* rubout - from current location, draw a rectangle in the backing color 
	obliterate any characters there */
static int
rubout( c, n )
int c; /* 1st character to rub out (first is 0) */
int n; /* number of characters to obliterate */
{
double rx1, ry1, rx2, ry2;

Ecblock( x+(c*Ecurtextwidth), y-MARG, 
	x+((c+n)*Ecurtextwidth), y+height+MARG, bkcolor, 0 );
return( 0 );
}

/* -------------------- */
/* cursor - draw (erase) the cursor to point to position (col) c */
static int
cursor( c, mode )
int c;
int mode; /* 1 if draw, 0 if erase */
{
double xpos;
xpos = (x + ( (c) * Ecurtextwidth)) - 0.02;
if( mode == 0 ) Ecolor( bkcolor );
Emov( xpos, y-0.02 ); Elin( xpos, y+height );
Emov( xpos-0.03, y-0.02 ); Elin( xpos+0.03, y-0.02 );
Emov( xpos-0.03, y+height ); Elin( xpos+0.03, y+height );
if( mode == 0 ) Ecolor( "black" );
return( 0 );
}

/* ----------------------- */
/* --- for multi-line ---- */
/* ----------------------- */

/* ----------------------- */
/* MAKEBREAKLIST - Make a list of all line breaks based on presence 
	of hard returns and calculated soft returns. */
static int
makebreaklist( txt, width )
char *txt;
int width;
{
int i, j, k;
int txtlen;
int lastb;
int lastspace;

Break[0] = -1;
Nlines = 1;
txtlen = strlen( txt );

lastb = 0;
lastspace = 0;
for( i = 0; i < strlen( txt ); i++ ) {

	if( txt[i] == '\n' ) {
		Break[Nlines++] = i;
		lastb = i;
		continue;
		}
	if( txt[i-1] == ' ' && txt[i] != ' ' ) lastspace = i-1;

	if( i - lastb > width || (lastb==0 && i-lastb>=width) ) {
		if( lastspace < lastb || ( lastspace == 0 && lastb == 0 )) { 
			Alignerrflag = 1; 
			return( 0 ); 
			}

		Break[Nlines++] = lastspace;	
		lastb = lastspace;
		i = lastspace;
		continue;
		}
	}
/* printf( "[makebreaklist: txtlen=%d  nlines=%d ]", strlen( txt ), Nlines ); */

/* last break is end of the text buffer.. */
Break[Nlines] = strlen( txt );
return( 0 );
}

/* -------------------- */
/* REFRESH */
/* -------------------- */
static int
refresh( buf, colwidth, width, nlines )
char *buf;
int colwidth;
double width;
int nlines;
{
int i, j;
char oldchar;
int stop;
Ecblock( x-MARG, (y-height)-MARG, x+width+MARG, y+MARG, bkcolor, 0 );
Ecblockdress( x-MARG, (y-height)-MARG, x+width+MARG, y+MARG, 
	0.02, "gray(1.0)", "gray(0.3)", 0.0, "" );

Alignerrflag = 0;

/* find line breaks.. */
makebreaklist( buf, colwidth );

if( Alignerrflag ) return( 0 );

/* display.. */
for( i = 0; i < Nlines; i++ ) {
	if( i >= nlines ) break; /* don't overflow bottom of box.. */
	if( i < (Nlines-1) ) stop = Break[i+1];
	else stop = strlen( buf );
	oldchar = buf[stop];
	buf[stop] = '\0';
	stpp( Break[i]+1 );
	Etext( &buf[ Break[i]+1 ] );
	buf[stop] = oldchar;
	}
/* for( i = 0; i < strlen( buf ); i++ ) {
 *	printf( "%c", buf[i] );
 *	for( j = 0; j < Nlines; j++ ) { if( i == Break[j] ) { printf( "$" ); break; } }
 *	}
 * fflush( stdout );
 */
 
}

/* ------------- */
/* STPP - set current text position for multi-line editing */
static int
stpp( k )
int k; /* an index into buffer */
{
int i;
int line;
double px, py;

/* figure out what line k is on.. */
for( i = 0; i < Nlines; i++ ) if( k <= Break[i] ) break;
line = i;
py = y - (line * Ecurtextheight);
if( line > 0 ) px = x + ((k - (Break[line-1]+1) )*Ecurtextwidth);
else px = x + (k*Ecurtextwidth);
Emov( px, py );
return( 0 );
}
/* ------------- */
/* CURSORPP - draw/erase cursor at x,y location corresponding to buf[k] */
static int
cursorpp( k, mode )
int k, mode;
{
double px, py;
double curtop;
int i, line;

/* figure out what line k is on and find y location.. */
for( i = 0; i < Nlines; i++ ) if( k <= Break[i] ) break;
line = i;
py = y - (line * Ecurtextheight);
/* figure out x location */
if( line > 0 ) px = x + ((k - (Break[line-1]+1))*Ecurtextwidth);
else px = x + (k*Ecurtextwidth);

/* printf( "[row=%d, col=%d]\n", line, k - (Break[line-1]+1) ); fflush( stdout ); */

if( mode == 0 ) Ecolor( bkcolor );
curtop = py + Ecurtextheight;
Emov( px, py-0.02 ); Elin( px, curtop );
Emov( px-0.03, py-0.02 ); Elin( px+0.03, py-0.02 );
Emov( px-0.03, curtop ); Elin( px+0.03, curtop );
if( mode == 0 ) Ecolor( "black" );
}

/* --------------- */
/* DO_UPDOWN - handle requests to move cursor up or down */
static int
do_updown( c, key, colhint )
int *c; /* current buffer location; updated */
int key; /* the arrow key that was pressed, either UP or DOWN */
int *colhint; /* remember current column even when moving onto a short line */
{
int i, k, line, col;

k = *c;

/* given a buffer position, find corresponding line and col.. */
for( line = 0; line < Nlines; line++ ) if( k <= Break[line] ) break;
if( line > 0 ) col = (k - (Break[line-1]+1));
else col = k;

/* adjust if there's a colhint.. */
if( col < *colhint ) col = *colhint;


/* see what's directly up/down from current position.. */
if( key == UP && line <= 1 ) return( 0 ); 
if( key == DOWN && line >= Nlines ) return( 0 ); 

/* update 'line' */
if( key == UP ) line--;
if( key == DOWN ) line++;

	
/* given a line and col, find corresponding buffer position.. */
if( line > 0 ) {

	/* if past end of line reposition at end of line */
	if( ( Break[line-1] + col ) > Break[line] )  k = Break[line];

	else k = ( Break[line-1] + col ) + 1;
	}
else k = col;

/* printf( "[new: line=%d col=%d k=%d]\n", line, col, k ); fflush( stdout ); */

*colhint = col;

*c = k;
return( 0 );
}

/* ------------------------------------- */
/* DO_AE - handle ^A (HOME) and ^E (END) */
static int
do_ae( c, key )
int *c;
int key;
{
int line;

for( line = 0; line < Nlines; line++ ) if( (*c) <= Break[line] ) break;
if( key == HOME || key == CTRL_A ) {
	if( line <= 1 ) *c = 0;
	else *c = Break[line-1] + 1;
	}
else if( key == END || key == CTRL_E ) {
	*c = Break[line];
	}
return( 0 );
}
/* ------------------------------------- */
/* DO_MOUSE - given a mouse click at x, y find equivalent buffer position */
static int
do_mouse( px, py, c )
double px, py;
int *c;
{
int line, col;
double y_lo;

line = ((y - py) / Ecurtextheight) + 1;
col = (px - x) / Ecurtextwidth;
/* printf( "[line:%d col:%d]\n", line, col ); fflush( stdout ); */ 
if( line > Nlines ) line = Nlines;
if( line > 0 ) {

	/* if past end of line reposition at end of line */
	if( ( Break[line-1] + col ) > Break[line] )  *c = Break[line];

	else *c = ( Break[line-1] + col ) + 1;
	}
else *c = col;
return( 0 );
}
/* ------------------------------------- */
/* PROCESS_NEWLINES - */
static int
process_newlines( buf, mode )
char *buf, *mode;
{
int i;
if( mode[0] == 'a' ) return( 0 ); /* asis */

else if( mode[0] == 'e' ) {  /* everyline */
	if( strlen( buf )==0 ) return( 0 ); 
	for( i = 0; i < Nlines; i++ ) buf[ Break[i] ] = '\n';
	buf[ Break[Nlines] ] = '\n';
	buf[ Break[Nlines] + 1 ] = '\0';
	return( 0 );
	}

else if( mode[0] == 'n' ) {  /* none */
	for( i = 0; i < strlen( buf ); i++ ) {
		if( buf[i] == '\n' ) buf[i] = ' ';
		}
	return( 0 );
	}
}


/* ------------------------------------- */
/* GETGUI_EDITPP - allow interactive entry/editing of multi-line text */

getgui_editpp( mode, buf, x1, y1, buflen, colwidth, pos, modflag, newlinemode )
int mode; /* E = active   D = display only (inactive)  R = redraw as active */
char *buf;
double x1, y1; /* location of upper left corner */
int buflen;
int colwidth;
int *pos;
int *modflag;
char *newlinemode; /* how to process results with respect to embedded newlines;
			 either "asis", "none", or "everyline" */
{
double width;
double ex, ey;
int key;
int c;
int nlines;
int colhint; 
int dlindex;
int textsize;
char *str;
int permmode;

/* printf( "editpp: mode=%c buf=%s pos=%g,%g len=%d colwidth=%d pos=%d\n",
	mode, buf, x1, y1, buflen, colwidth, *pos ); fflush( stdout ); */

if( mode == 'D' ) strcpy( bkcolor, Color3 );
else strcpy( bkcolor, Color4 );

nlines = buflen / colwidth;

x = x1;
y = y1;
height = nlines * Ecurtextheight;
width = colwidth * Ecurtextwidth;
colhint = 0;
textsize = Ecurtextsize;

refresh( buf, colwidth, width, nlines );

/* position cursor at 0 */
if( *pos >= strlen( buf )) c = strlen( buf );
else c = *pos;
if( mode != 'D' ) cursorpp( c, 1 );

/* save info in display list (use 'permanent' mode) */
if( mode == 'R' ) permmode = 'E';
else permmode = mode;
if( Keepdl ) dlindex = getgui_dladd( "M", buf, x-MARG, (y-height)-MARG, x+width+MARG, y+MARG, 
		x, y, permmode, buflen, *pos, Ecurtextsize, colwidth );

/* find this editln's location in the display list */
else dlindex = getgui_dlfind( x, y );

if( mode == 'D' || mode == 'R' ) return( 0 ); /* redraws return here.. */

/* work directly on entry in display list, in case an expose occurs in the
   midst of what we are doing.. */
str = &DLbigbuf[ atoi( DLval[ dlindex ] ) ];


while( 1 ) {
        DLi3[ dlindex ] = c; 
	Egetkey( &ex, &ey, &key );
	/* printf( "===== [key=%d]", key ); fflush( stdout ); */

        /* make sure textsize hasn't changed underfoot by expose/redraw */
        if( Ecurtextsize != textsize ) Etextsize( textsize );


	if( key != UP && key != DOWN ) colhint = 0;

	/* mouse buttons */
	if( key > 1000 ) {
		if( ex < x || ey < y-height || ex > x + width || ey > y ) {
			strcpy( buf, str );
			process_newlines( buf, newlinemode );
			return( 0 );
			}
		do_mouse( ex, ey, &c );
		}

       if( key == LEFT ) {
       		if( c > 0 ) { 
			cursorpp( c, 0 ); c--; cursorpp( c, 1 ); 
			continue; 
			}
               	}
 
       else if( key == RIGHT ) {
               	if(  c < strlen( str ) ) { 
			cursorpp( c, 0), c++; cursorpp( c, 1 ); 
			continue;
			}
               	}

	else if( key == UP || key == DOWN ) {
		cursorpp( c, 0 );
		do_updown( &c, key, &colhint );
		cursorpp( c, 1 );
		continue;
		}

	else if( key == CTRL_A || key == CTRL_E || key == HOME || key == END ) { 
		cursorpp( c, 0 ); 
		do_ae( &c, key ); 
		cursorpp( c, 1 );
		continue;
		}

	else if( key == DEL || key == BACKSPACE || key == CTRL_D ) {
		/* because of the cursor shape, there is no concept of 'DELETE' */
		*modflag = 1;
		if( c == 0 ) continue;
		c--;
		delbuf( str, c, 1 );
		}

	else if( key == CTRL_K ) {  /* delete to EOL */
		int stop;
		stop = c;
		do_ae( &stop, CTRL_E );
		delbuf( str, c, (stop-c) );
		*modflag = 1;
		}

	else if( key == CTRL_U ) {  /* delete line */
		int start, stop;
		start = stop = c;
		do_ae( &start, CTRL_A );
		do_ae( &stop, CTRL_E );
		if( start == 0 ) delbuf( str, start, (stop-start) );
		else delbuf( str, start-1, (stop-start)+1 );
		if( start == 0 ) c = 0;
		else c = start;
		*modflag = 1;
		}


	/* typed characters.. */
	else if( (key >= 32 && key <= 126) ) {
		if( strlen( str ) >= buflen ) { printf( "" ); fflush( stdout ); continue; }
		*modflag = 1;
		addbuf( str, c, key );
		c++;
		}

	else if( key == 13 ) {
		if( strlen( str ) >= buflen ) { printf( "" ); fflush( stdout ); continue; }
		*modflag = 1;
		addbuf( str, c, '\n' );
		c++;
		}

	refresh( str, colwidth, width, nlines );

	/* check for too many lines.. */
	if( Nlines > nlines || Alignerrflag ) {
		/* delete most recently added character.. */
		c--;
		delbuf( str, c, 1 );
		/* beep? */
		printf( "" ); fflush( stdout );
		refresh( str, colwidth, width, nlines );
		}

	cursorpp( c, 1 );
	}

/* I don't think control will ever reach here.. */
process_newlines( buf, newlinemode );
return( 0 ); 
}
